#include <linux/module.h>
#include <linux/kernel.h>
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <rtl_sched.h>
#include <asm/io.h>
#include <linux/cons.h>
#include "common.h"

int number_of_tests=500;
int period=2000000;
int periodic_mode=0;
int fifo_size=4000;

pthread_t thread;
timer_t timer;
struct timespec expected;

/* For testing purposes, if we are on a smp machine, try to 
  run the measurement module on the _other_cpu 
  on a uniprocessor, ignore this test.
*/
  
#ifdef __SMP__
void test_change_cpus(void){
	DECLARE_CPUID(cpu_id);
	int assign_status;
	int newcpu = !cpu_id;
	rtl_printf("thread executes on cpu %d\n", cpu_id);
	rtl_printf("assigning thread to cpu %d\n", newcpu);

	assign_status = rtl_pthread_assign_to_cpu (thread, newcpu);

	if (assign_status < 0) {
		rtl_printf("failed\n");
	} else {
		cpu_id = rtl_get_cpu_id();
		rtl_printf("now thread executes on cpu %d\n", cpu_id);
	}
}
#else 
#define test_change_cpus()  
#endif

void *thread_code(void *t) {

	struct timespec diff;
	struct timespec now;
	struct timespec min_diff;
	struct timespec max_diff;
	struct sample samp;
	struct itimerspec spec;
	int i;
	int cnt = 0;

	clock_gettime (CLOCK_REALTIME, &spec.it_value);	
	spec.it_interval = timespec_from_ns (period);

	timer_settime (timer, TIMER_ABSTIME, &spec, NULL);
	expected = spec.it_value;

	while (1) {
		min_diff = timespec_from_ns(2000000000);
		max_diff = timespec_from_ns(-2000000000);

		for (i = 0; i < number_of_tests; i++) {
/* 			conpr("t"); */
			++cnt;
			if (cnt == 1000) {
				test_change_cpus();
			}
			rtl_pthread_wait_timer(timer);

			clock_gettime (CLOCK_REALTIME, &now);
			diff = now;
			timespec_sub (&diff, &expected);

			timespec_add (&expected, &spec.it_interval);

			if (timespec_lt (&diff, &min_diff)) {
				min_diff = diff;
			}
			if (timespec_gt (&diff, &max_diff)) {
				max_diff = diff;
			}

		}

		samp.min = min_diff;
		samp.max = max_diff;
		rtf_put(0, &samp, sizeof(samp));
/* 		rtl_printf("rmin: %11d, rmax: %11d\n", (int) (timespec_to_ns(&samp.min)), (int) (timespec_to_ns(&samp.max))); */
	}
}



int init_module(void)
{
	pthread_attr_t attr;
	struct rtl_sched_param sched_param;
	int thread_status;
        int fifo_status;
/* 	rtl_debug(RTL_DEBUG_TIMER); */

	fifo_status = rtf_create(0, fifo_size);
	if ((fifo_status != fifo_size)) {
		printk("RTL measurement test fail. fifo_status=%d\n",fifo_status);
		return -1;
	}

/* 	rtf_put(0, &fifo_status, sizeof(fifo_status)); */

	timer_create(CLOCK_REALTIME, NULL, &timer);

	printk("RTL measurement module on CPU %d\n",rtl_get_cpu_id());
	pthread_attr_init (&attr);
	sched_param.sched_priority = 1;
	pthread_attr_setschedparam (&attr, &sched_param);
	thread_status = pthread_create (&thread,  &attr, thread_code, (void *)1);
	if (thread_status < 0) {
		printk("failed to create RT-thread\n");
		return -1;
	} else {
		printk("created RT-thread\n");
	}

	return 0;
}


void cleanup_module(void)
{
        printk ("Removing module on CPU %d\n", rtl_get_cpu_id());
	rtl_pthread_delete (thread);
/* should be:	pthread_cancel(thread); pthread_join(thread); */
/* or:  	rtl_pthread_kill_other_threads(); */
	timer_delete(timer);
	rtf_destroy(0);
}

