/*
 * rtl_time.c
 *
 * architecture-dependent clock support
 *
 * Written by Michael Barabanov
 * Copyright VJY Associates 1998-1999
 * Released under the terms of the LGPL
 *
 */

#define RTL_TIME_DEBUG

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <linux/errno.h>
#include <asm/system.h>
#include <linux/rtl.h>
#include <linux/sched.h>
#include <linux/timex.h>
#include <linux/cons.h>
#include <linux/mc146818rtc.h>
#include <rtl_sync.h>
#include <rtl_time.h>

/* struct clock_struct t; */
/* int rtl_time_dummy_aligner __attribute__ ((aligned(128))); */

extern void (*kd_mksound)(unsigned int hz, unsigned int ticks);

static long long (*rtl_do_get_time)(void);

static long LATCH_NS;
static struct timespec LATCH_TIMESPEC;
static struct timespec MAX_LATCH_ONESHOT;

/* getting global time from 8254 */

#define READ_CNT0(var) \
{ var = inb(0x40); var |= (inb(0x40) << 8); }

#define READ_CNT2(var) \
{ var = inb(0x42); var |= (inb(0x42) << 8); }


#define LATCH_CNT0() \
outb(0xd2, 0x43);

#define LATCH_CNT0_AND_2() \
outb(0xda, 0x43);

#define LATCH_CNT2() \
outb(0xd8, 0x43);

#define WRITE_COUNTER_ZERO16(x)  { \
	outb_p(x&0xff,0x40); outb_p((x>>8)&0xff,0x40);\
      	clock_counter =x; \
} 

#define WRITE_COUNTER_ZERO8(x)  { \
	outb_p(x&0xff,0x40); \
	clock_counter =x; \
}

#define WRITE_COUNTER_ZERO_ONESHOT(x) { \
	if (x <= 0xff) { \
		if (max_counter0 != 0xff) { \
			outb_p(0x10, 0x43);       /* 8254, channel 0, mode 0, lsb */ \
			max_counter0 = 0xff; \
		} \
		WRITE_COUNTER_ZERO8(x); \
	} else { \
		if (max_counter0 != 0xffff) { \
			outb_p(0x30, 0x43);        /* 8254, channel 0, mode 0, lsb+msb */ \
			max_counter0 = 0xffff; \
		} \
		WRITE_COUNTER_ZERO16(x); \
	} \
}

static volatile long long rt_time; /* last time (in 8254 ticks) when counter 2 was equal to last_c2 */
static volatile int last_c2;
unsigned long scaler_8254_to_ns;
#define LATCH2 0x8000

#ifdef __SMP__
static spinlock_t lock8254;
/* static spinlock_t rtl_time_lock; */
#define RTL_SPIN_LOCK rtl_time_lock
#endif

clockid_t clock_local[NR_CPUS];

long long global_8254_gettime (void)
{
	register int c2;
	int flags;
	long long t;

	rtl_spin_lock_irqsave (&lock8254, flags);
	LATCH_CNT2();
	READ_CNT2(c2);
	rt_time += ((c2 < last_c2) ? (last_c2 - c2) / 2 : (last_c2 - c2 + LATCH2) / 2);
	last_c2 = c2;
	__asm__(
		"shld $10, %%eax, %%edx\n\t"
		"shl $10, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t) : "b" (scaler_8254_to_ns), "A" (rt_time): "cx");
	rtl_spin_unlock_irqrestore (&lock8254, flags);
	return t;
}


/* getting global time from Pentium TSC */
static unsigned long scaler_pentium_to_ns;
extern unsigned long fast_gettimeoffset_quotient;
long rtl_delay_mult;
int can_change_latch2;

long long pent_gettime(void)
{

	long long t;

	/* time = counter * scaler_pentium_to_ns / 2^32 * 2^5; */
	/* Why 2^5? Because the slowest Pentiums run at 60 MHz */

	__asm__("rdtsc\n\t"
		"mov %%edx, %%ecx\n\t"
		"mul %%ebx\n\t"  	/* multiply the low 32 bits of the counter by the scaler_pentium */
		"mov %%ecx, %%eax\n\t"
		"mov %%edx, %%ecx\n\t"	/* save the high 32 bits of the product */
		"mul %%ebx\n\t" 	/* now the high 32 bits of the counter */
		"add %%ecx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		"shld $5, %%eax, %%edx\n\t"
		"shl $5, %%eax\n\t"
		:"=A" (t) : "b" (scaler_pentium_to_ns) : "cx");
	return t;
}

static unsigned long (*save_do_gettimeoffset)(void);
/* we can't allow Linux to read the clocks; it must use this function */
static unsigned long do_rt_gettimeoffset(void)
{
	return 0; /* TODO: take code from 2.0 version */
}


static void rtl_kd_nosound(unsigned long ignored)
{
	int flags;
	rtl_no_interrupts(flags);
	outb(inb_p(0x61) & 0xfd, 0x61);
	rtl_restore_interrupts(flags);
}

static void rtl_kd_mksound(unsigned int hz, unsigned int ticks)
{
	static struct timer_list sound_timer = { NULL, NULL, 0, 0,
						 rtl_kd_nosound };

	unsigned int count = 0;

	if (hz > 20 && hz < 32767)
		count = 1193180 / hz;
	
	cli();
	del_timer(&sound_timer);
	if (count) {
		int flags;
		rtl_spin_lock_irqsave (&lock8254, flags);
		outb_p(inb_p(0x61)|3, 0x61);
		if (can_change_latch2) {
			outb_p(0xB6, 0x43);
			outb_p(count & 0xff, 0x42);
			outb((count >> 8) & 0xff, 0x42);
		}
		rtl_spin_unlock_irqrestore (&lock8254, flags);

		if (ticks) {
			sound_timer.expires = jiffies+ticks;
			add_timer(&sound_timer);
		}
	} else
		rtl_kd_nosound(0);
	sti();
	return;
}


static void (*save_kd_mksound)(unsigned int hz, unsigned int ticks);


static int clock_realtime_init (clockid_t clock)
{
	int flags;
/* 	struct timespec current_time; */

	kd_mksound(0, 0); /* clear the possibly pending sound timer */
	save_kd_mksound = kd_mksound;
	kd_mksound = rtl_kd_mksound;

	scaler_8254_to_ns = muldiv (NSECS_PER_SEC, 1 << 22, CLOCK_TICK_RATE);
	LATCH_NS = muldiv (LATCH, NSECS_PER_SEC, CLOCK_TICK_RATE);
	LATCH_TIMESPEC.tv_sec = 0;
	LATCH_TIMESPEC.tv_nsec = LATCH_NS;

	MAX_LATCH_ONESHOT.tv_sec = 0;
	MAX_LATCH_ONESHOT.tv_nsec = LATCH_NS * 3 / 4;

	rtl_no_interrupts(flags);
	if (boot_cpu_data.x86_capability & X86_FEATURE_TSC) {
		can_change_latch2 = 1;
		/* scaler_pentium ==  2^32 / (2^5 * (cpu clocks per ns)) */

		scaler_pentium_to_ns = muldiv (get_tsc_quotient(), 1000, 1 << 5);
/* 		printk("scaler_pentium_to_ns = %d\n", (int) scaler_pentium_to_ns); */
		rtl_do_get_time = pent_gettime;
	} else {
		can_change_latch2 = 0;
		/* program channel 2 of the 8254 chip for periodic counting */
		outb_p(0xb6, 0x43);     /* binary, mode 3, LSB/MSB, ch 2 */
		outb_p(LATCH2 & 0xff, 0x42);
		outb_p((LATCH2 >> 8) & 0xff, 0x42);
		outb_p((inb_p(0x61) & 0xfd) | 1, 0x61); /* shut up the speaker and enable counting */

		LATCH_CNT2();
		READ_CNT2(last_c2);
		rt_time = 0;
		rtl_do_get_time = global_8254_gettime;
	}


	save_do_gettimeoffset = do_gettimeoffset;
	do_gettimeoffset = do_rt_gettimeoffset;

	/*
	current_time.tv_sec = xtime.tv_sec;
	current_time.tv_nsec = 0;

	rtl_time_offset = 0;
	rtl_time_offset = rtl_timespec_to_rtime(&current_time) - rtl_get_time();

	*/
	rtl_restore_interrupts(flags);
	return 0;
}

static void clock_realtime_uninit (clockid_t clock)
{
	do_gettimeoffset = save_do_gettimeoffset;
	rtl_kd_mksound(0, 0);
	kd_mksound = save_kd_mksound;
}


int rtl_dummy_clock_init (clockid_t clock)
{
	return -EINVAL;
}

void rtl_dummy_clock_checklinuxirq (clockid_t clock)
{
}

void rtl_dummy_clock_uninit (clockid_t clock)
{
}

int rtl_dummy_clock_gettime (clockid_t clock, struct timespec *ts)
{
	return -EINVAL;
}

int rtl_dummy_clock_settimer (clockid_t clock, int mode, struct timespec *interval) {
	return -EINVAL;
}

static int clock_realtime_gettime (clockid_t clock, struct timespec *ts)
{
	long long t = rtl_do_get_time();
	__asm__("divl %%ecx\n\t"
		:"=a" (ts->tv_sec), "=d" (ts->tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	return 0;
}



static void default_handler(clockid_t clock, struct pt_regs *r)
{
	struct timespec ts;
	clock_gettime (CLOCK_REALTIME, &ts);
	rtl_clock_settimer (clock, rtl_clock_getmode(clock), NULL);
	rtl_clock_checklinuxirq(clock);
}


/* the 8254 clock */

static unsigned int clock_counter; /* current latch value */
static int max_counter0;    /* 0xff or 0xffff - one or two byte counts */

static struct timespec rt_linux_clock;/* time to generate the next linux interrupt  */

unsigned long scaler_ns_to_8254;

static inline long RTIME_to_8254_ticks(long t)
{
	__asm__("mull %0"
		:"=d" (t)
		:"a" (t), "0" (scaler_ns_to_8254)
		:"ax");
		
	return (t);
}


static unsigned int _8254_irq(unsigned int irq, struct pt_regs *regs)
{
	int flags;
	rtl_spin_lock_irqsave (&lock8254, flags);
	if (clock_data(CLOCK_8254)->mode == CLOCK_MODE_PERIODIC) {
		timespec_add (&clock_data(CLOCK_8254)->time, &clock_data(CLOCK_8254)->resolution);
	} else {
		clock_data(CLOCK_8254)->istimerset = 0;
	}
	rtl_spin_unlock_irqrestore (&lock8254, flags);
	clock_data(CLOCK_8254) -> handler (CLOCK_8254, regs);

	return 0;
}


static int _8254_settimer (clockid_t clock, int mode, struct timespec *interval)
{
	int flags;
	long t;
	struct timespec ts;

	rtl_spin_lock_irqsave (&lock8254, flags);
	if (mode == CLOCK_MODE_PERIODIC && interval) {
		if (clock_data(clock)->mode != mode) {
			outb_p(0x34,0x43); /* binary, mode 2, LSB/MSB, ch 0 */
			clock_data(clock) -> mode = mode;
		}
		t = RTIME_to_8254_ticks (interval -> tv_nsec);
		WRITE_COUNTER_ZERO16 (t);

		clock_gettime (CLOCK_REALTIME, &clock_data(clock)->time);
		clock_data(clock) -> resolution.tv_sec = 0;
		clock_data(clock) -> resolution.tv_nsec = muldiv (t, NSECS_PER_SEC, CLOCK_TICK_RATE);
		clock_data(CLOCK_8254)->istimerset = 1;
	} else if (mode == CLOCK_MODE_ONESHOT) {
		if (clock_data(clock)->mode != mode) {
			outb_p(0x30, 0x43);       /* 8254, channel 0, mode 0, lsb+msb */
			max_counter0 = 0xffff;
			clock_data(clock) -> mode = mode;
			clock_data(clock) -> resolution.tv_sec = 0;
			clock_data(clock) -> resolution.tv_nsec = 1;
		}

		if (interval) {
			ts = *interval;
			if (timespec_gt(&ts, &MAX_LATCH_ONESHOT)) {
				ts = MAX_LATCH_ONESHOT;
			} else if (timespec_lt(&ts, &clock_data(clock)->resolution)) {
				ts = clock_data(clock)->resolution;
			}
		} else {
			ts = MAX_LATCH_ONESHOT;
		}

		t = RTIME_to_8254_ticks(ts.tv_nsec) + 1; /* FIXME (+1) */
		WRITE_COUNTER_ZERO_ONESHOT(t);
		clock_data(CLOCK_8254)->istimerset = 1;
	}

	rtl_spin_unlock_irqrestore(&lock8254, flags);
	return 0;
}


/* #define __RTL_DEBUG_TIME__ */

void _8254_checklinuxirq (clockid_t clock)
{
/*	if (test_and_set_bit(0, &checklinuxirq[clock])) {
		return;
	}
	*/
	struct timespec ts;

	clock_gettime (CLOCK_REALTIME, &ts);
	if  (timespec_ge (&ts, &rt_linux_clock)) {
		timespec_add (&rt_linux_clock, &LATCH_TIMESPEC);
		rtl_global_pend_irq (0);
		if  (timespec_ge (&ts, &rt_linux_clock)) {
#ifdef __RTL_DEBUG_TIME__
			conpr("RTL: lost jiffies\n");
#endif
			rt_linux_clock = ts;
			timespec_add (&rt_linux_clock, &MAX_LATCH_ONESHOT);
		}
	}

	if (!clock_data(CLOCK_8254)->istimerset) {
		_8254_settimer (CLOCK_8254, CLOCK_MODE_ONESHOT, &MAX_LATCH_ONESHOT);
	}
/* 	clear_bit (0, &checklinuxirq [clock]); */
}

static int periodic_clock_gettime (clockid_t clock, struct timespec *ts)
{
	if (clock_data(clock)->mode == CLOCK_MODE_PERIODIC) {
		clock_gettime (CLOCK_REALTIME, ts);
	} else {
		*ts = clock_data(clock)->time;
	}
	return 0;
}


static int _8254_init (clockid_t clock)
{
	int flags;

	scaler_ns_to_8254 = muldiv (CLOCK_TICK_RATE, 1 << 31, NSECS_PER_SEC / 2);
	rtl_no_interrupts (flags);

	clock_gettime(CLOCK_REALTIME, &rt_linux_clock);
	timespec_add (&rt_linux_clock, &LATCH_TIMESPEC);

	rtl_request_global_irq(0, _8254_irq);
	_8254_settimer (clock, CLOCK_MODE_PERIODIC, &LATCH_TIMESPEC);

	rtl_restore_interrupts (flags);
	return 0;
}

static void _8254_uninit (clockid_t clock)
{
	int flags;

	rtl_spin_lock_irqsave (&lock8254, flags);

	outb_p(0x34,0x43);		/* binary, mode 2, LSB/MSB, ch 0 */
	WRITE_COUNTER_ZERO16(LATCH);
	rtl_free_global_irq(0);

	rtl_spin_unlock_irqrestore (&lock8254, flags);
}



#ifdef __SMP__
/* APIC clocks */

#define LOCAL_TIMER_VECTOR	0x41

static unsigned long scaler_ns_to_apic;

struct clock_per_cpu_struct {
	struct timespec linux_apic;
	char pad[64];
} rt_linux_clock_smp[NR_CPUS];		/* time to generate the next linux interrupt  */

static inline long RTIME_to_apic_ticks(long t)
{
	__asm__("mull %0"
		:"=d" (t)
		:"a" (t), "0" (scaler_ns_to_apic)
		:"ax");
		
	return (t);
}


static inline int rtl_apic_write_initial_count (long count)
{
	unsigned int tmp_value;
	tmp_value = apic_read(APIC_TMICT);
	apic_write (APIC_TMICT, count);
	return 0;
}

/* this is defined in arch/i386/kernel/smp.c */
#define APIC_DIVISOR 16

static int apic_settimer (clockid_t clock, int mode, struct timespec *interval)
{
	int flags;
	long t;
	struct timespec ts;
	unsigned long lvtt1_value;
	unsigned int tmp_value;
	DECLARE_CPUID(cpu_id);

#ifdef RTL_TIME_DEBUG
		if (clock - CLOCK_APIC != cpu_id) {
			conpr("RTL panic: apic_settimer called on wrong CPU!\n");
			rtl_no_interrupts (flags);
			while (1);
		}
#endif

	rtl_no_interrupts (flags);
	if (mode == CLOCK_MODE_PERIODIC && interval) {
		if (clock_data(clock)->mode != mode) {
			clock_data(clock) -> mode = mode;
			tmp_value = apic_read(APIC_LVTT);
			lvtt1_value = APIC_LVT_TIMER_PERIODIC | LOCAL_TIMER_VECTOR;
			apic_write(APIC_LVTT , lvtt1_value);
		}
		t = RTIME_to_apic_ticks(interval->tv_nsec);
		rtl_apic_write_initial_count (t);
		clock_gettime (CLOCK_REALTIME, &clock_data(clock)->time);
		clock_data(clock) -> resolution = *interval;
		clock_data(clock)->istimerset = 1;
	} else if (mode == CLOCK_MODE_ONESHOT) {
		if (clock_data(clock)->mode != mode) {
			clock_data(clock) -> mode = mode;
			clock_data(clock) -> resolution.tv_sec = 0;
			clock_data(clock) -> resolution.tv_nsec = 1;

			tmp_value = apic_read(APIC_LVTT);
			lvtt1_value = LOCAL_TIMER_VECTOR;
			apic_write(APIC_LVTT , lvtt1_value);

		}

		if (interval) {
			ts = *interval;
			if (timespec_gt(&ts, &MAX_LATCH_ONESHOT)) {
				ts = MAX_LATCH_ONESHOT;
			} else if (timespec_lt(&ts, &clock_data(clock)->resolution)) {
				ts = clock_data(clock)->resolution;
			}
		} else {
			ts = MAX_LATCH_ONESHOT;
		}


		t = RTIME_to_apic_ticks(ts.tv_nsec) + 1; /*FIXME (+1)*/
		rtl_apic_write_initial_count (t);
		clock_data(clock)->istimerset = 1;
	}

	rtl_restore_interrupts(flags);
	return 0;
}

void apic_checklinuxirq (clockid_t clock)
{
	DECLARE_CPUID(cpu_id);
	struct timespec ts;
	clock_gettime (CLOCK_REALTIME, &ts);
	if  (timespec_ge (&ts, &rt_linux_clock_smp[cpu_id].linux_apic)) {
		rtl_local_pend_vec (LOCAL_TIMER_VECTOR,cpu_id);
		timespec_add (&rt_linux_clock_smp [cpu_id].linux_apic, &LATCH_TIMESPEC);
		if  (timespec_ge (&ts, &rt_linux_clock_smp[cpu_id].linux_apic)) {
#ifdef __RTL_DEBUG_TIME__
			conpr("RTL: lost linux APIC timer interrupts\n");
#endif
			rt_linux_clock_smp[cpu_id].linux_apic = ts;
			timespec_add (&rt_linux_clock_smp[cpu_id].linux_apic, &MAX_LATCH_ONESHOT);
		}
	}
	if (!clock_data (CLOCK_APIC + cpu_id) -> istimerset) {
		apic_settimer (CLOCK_APIC + cpu_id, CLOCK_MODE_ONESHOT, &MAX_LATCH_ONESHOT);
	}
}

static unsigned int apic_timer_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	if (clock_data(CLOCK_APIC + cpu_id)->mode == CLOCK_MODE_PERIODIC) {
		timespec_add (&clock_data(CLOCK_APIC + cpu_id)->time, &clock_data(CLOCK_APIC + cpu_id)->resolution);
	} else {
		clock_data(CLOCK_APIC + cpu_id)->istimerset = 0;
	}
	clock_data(CLOCK_APIC + cpu_id) -> handler (CLOCK_APIC + cpu_id, r);
	return 0;
}


int volatile apic_init_flag;

static unsigned int apic_init_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	apic_settimer (cpu_id + CLOCK_APIC, CLOCK_MODE_ONESHOT, NULL);
	rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id);
	clear_bit (0, &apic_init_flag);
	return 0;
}

/* apics are initialized to the oneshot mode by default */
static int apic_clock_init (clockid_t clock)
{
	int flags;
	DECLARE_CPUID(cpu_id);

	scaler_ns_to_apic = muldiv (muldiv (get_apic_calibration_result(), 1 << 31, NSECS_PER_SEC / 2), CLOCK_TICK_RATE, LATCH * APIC_DIVISOR);

	rtl_no_interrupts (flags);
	clock_gettime(CLOCK_REALTIME, &rt_linux_clock_smp[clock - CLOCK_APIC].linux_apic);
	timespec_add (&rt_linux_clock_smp[clock - CLOCK_APIC].linux_apic, &LATCH_TIMESPEC);

	if (clock - CLOCK_APIC == cpu_id) {
		apic_settimer (clock, CLOCK_MODE_ONESHOT, NULL);
	} else {
		set_bit (0, &apic_init_flag);
		rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_init_irq, clock - CLOCK_APIC);
		while (test_bit(0, &apic_init_flag));
	}
	rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_timer_irq, clock - CLOCK_APIC);
	rtl_restore_interrupts (flags);
	return 0;
}


static unsigned int apic_uninit_irq(struct pt_regs *r)
{
	DECLARE_CPUID(cpu_id);
	apic_settimer (cpu_id + CLOCK_APIC, CLOCK_MODE_PERIODIC, &LATCH_TIMESPEC);
/* 	rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id); */
	free_ipi (cpu_id);
	clear_bit (0, &apic_init_flag);
	return 0;
}

static void apic_clock_uninit (clockid_t clock)
{
	int flags;
	DECLARE_CPUID(cpu_id);


	if (clock - CLOCK_APIC == cpu_id) {
		rtl_no_interrupts (flags);
		apic_settimer (cpu_id + CLOCK_APIC, CLOCK_MODE_PERIODIC, &LATCH_TIMESPEC);
		rtl_free_local_irq (LOCAL_TIMER_VECTOR, cpu_id);
		rtl_restore_interrupts (flags);
	} else {
#if 1
		rtl_no_interrupts (flags);

		set_bit (0, &apic_init_flag);
		rtl_free_local_irq (LOCAL_TIMER_VECTOR, clock - CLOCK_APIC);
		request_ipi (apic_uninit_irq, clock - CLOCK_APIC);
		rtl_reschedule(clock - CLOCK_APIC);
/* 		rtl_request_local_irq (LOCAL_TIMER_VECTOR, apic_uninit_irq, clock - CLOCK_APIC); */
		rtl_restore_interrupts (flags);
		while (test_bit(0, &apic_init_flag)); /* wait for apic to uninit */
#endif
	}

}


#endif



int clock_init (clockid_t clock)
{
	if (clock_data(clock) -> init (clock) < 0) {
		return -EIO;
	}
	return 0;
}


int rtl_clock_uninit (clockid_t clock)
{
	clock_data(clock) -> uninit (clock);
	return 0;
}

int rtl_clock_sethandler (clockid_t clock, RTL_CLOCK_HANDLER fn)
{
	if (clock_data(clock) -> handler != default_handler) {
		return -EBUSY;
	}
	clock_data(clock) -> handler = fn;
	return 0;
}

int rtl_clock_unsethandler (clockid_t clock)
{
	if (clock_data(clock) -> handler == default_handler) {
		return -EINVAL;
	}
	clock_data(clock) -> handler = default_handler;
	return 0;
}

int rtl_clock_ishandlerset (clockid_t clock)
{
	return clock_data(clock) -> handler != 0;
}


__asm__(".section .data\n\t"
	".align 64\n\t"
	".section .code\n\t");

#ifdef __SMP__
struct clock_struct apic_clock = {apic_clock_init, apic_clock_uninit, periodic_clock_gettime, apic_settimer, apic_checklinuxirq};
#endif

struct clock_struct clocks [RTL_NR_CLOCKS] /* __attribute__((section (".al64")))*/ = {
	{clock_realtime_init,
		clock_realtime_uninit,
		clock_realtime_gettime,
		rtl_dummy_clock_settimer,
		rtl_dummy_clock_checklinuxirq,
	},
	{rtl_dummy_clock_init, rtl_dummy_clock_uninit, rtl_dummy_clock_gettime, rtl_dummy_clock_settimer, rtl_dummy_clock_checklinuxirq},
	{_8254_init,
		_8254_uninit,
		periodic_clock_gettime,
		_8254_settimer,
		_8254_checklinuxirq,
	},
	{rtl_dummy_clock_init, rtl_dummy_clock_uninit, rtl_dummy_clock_gettime, rtl_dummy_clock_settimer, rtl_dummy_clock_checklinuxirq},
	{rtl_dummy_clock_init, rtl_dummy_clock_uninit, rtl_dummy_clock_gettime, rtl_dummy_clock_settimer, rtl_dummy_clock_checklinuxirq},
};



int init_module (void)
{
	int i;
/* 	DECLARE_CPUID(cpu_id); */
	spin_lock_init (&lock8254);
	

	rtl_delay_mult = muldiv (loops_per_sec, 1 << 31, NSECS_PER_SEC / 2);

	for (i = 0; i < NR_CPUS; i++) {
		clock_local[i] = CLOCK_8254;
	}

#ifdef __SMP__
	if (smp_found_config) {
		for (i = 0; i < 2; i++) {
			clocks[i + CLOCK_APIC] = apic_clock;
		}
		for (i = 0; i < NR_CPUS; i++) {
			clock_local[i] = CLOCK_APIC + i;
		}
	}
#endif

	for (i = 0; i < RTL_NR_CLOCKS; i++) {
		clocks [i].mode = CLOCK_MODE_UNINITIALIZED;
		clocks [i].handler = default_handler;
		clocks[i].istimerset = 0;
		clocks[i].init (i);
/* 		rtl_printf("%x\n", &clocks[i]); */
	}

/* 	rtl_clock_settimer (CLOCK_8254, CLOCK_MODE_ONESHOT, NULL); */

	return 0;
}

void cleanup_module(void)
{
	int i;
/* 	DECLARE_CPUID(cpu_id); */
	for (i = RTL_NR_CLOCKS - 1; i >= 0; i--) {
		clocks[i].uninit (i);
	}
#ifdef __SMP__
	if (smp_found_config) {
		/* SMP-specific cleanup */
	}
#endif

}

/* this really should belong to rtl_core */
#ifdef __SMP__
#define RTL_RESCHEDULE_VECTOR 0x51
int request_ipi (unsigned int (*f)(struct pt_regs *r), int cpu)
{
	return rtl_request_local_irq (RTL_RESCHEDULE_VECTOR, f, cpu);
}

int free_ipi (int cpu)
{
	return rtl_free_local_irq (RTL_RESCHEDULE_VECTOR, cpu);
}
#endif
